
# Copyright (c) 2011, 2012, 2013, 2015, 2016 Nicira, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at:
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime
import logging
import logging.handlers
import os
import re
import socket
import sys
import threading

import ovs.dirs
import ovs.unixctl
import ovs.util

DESTINATIONS = {"console": "info", "file": "info", "syslog": "info"}
PATTERNS = {
    "console": "%D{%Y-%m-%dT%H:%M:%SZ}|%05N|%c%T|%p|%m",
    "file": "%D{%Y-%m-%dT%H:%M:%S.###Z}|%05N|%c%T|%p|%m",
    "syslog": "ovs|%05N|%c%T|%p|%m",
}
LEVELS = {
    "dbg": logging.DEBUG,
    "info": logging.INFO,
    "warn": logging.WARNING,
    "err": logging.ERROR,
    "emer": logging.CRITICAL,
    "off": logging.CRITICAL
}
FACILITIES = ['auth', 'authpriv', 'cron', 'daemon', 'ftp', 'kern', 'lpr',
              'mail', 'news', 'syslog', 'user', 'uucp', 'local0', 'local1',
              'local2', 'local3', 'local4', 'local5', 'local6', 'local7']
syslog_facility = "daemon"
syslog_handler = ''


def get_level(level_str):
    return LEVELS.get(level_str.lower())


class Vlog(object):
    __inited = False
    __msg_num = 0
    __start_time = 0
    __mfl = {}  # Module -> destination -> level
    __log_file = None
    __file_handler = None
    __log_patterns = PATTERNS

    def __init__(self, name):
        """Creates a new Vlog object representing a module called 'name'.  The
        created Vlog object will do nothing until the Vlog.init() static method
        is called.  Once called, no more Vlog objects may be created."""

        assert not Vlog.__inited
        self.name = name.lower()
        if name not in Vlog.__mfl:
            Vlog.__mfl[self.name] = DESTINATIONS.copy()

    def __log(self, level, message, **kwargs):
        if not Vlog.__inited:
            return

        level_num = LEVELS.get(level.lower(), logging.DEBUG)
        msg_num = Vlog.__msg_num
        Vlog.__msg_num += 1

        for f, f_level in Vlog.__mfl[self.name].items():
            f_level = LEVELS.get(f_level, logging.CRITICAL)
            if level_num >= f_level:
                msg = self._build_message(message, f, level, msg_num)
                logging.getLogger(f).log(level_num, msg, **kwargs)

    def _build_message(self, message, destination, level, msg_num):
        pattern = self.__log_patterns[destination]
        tmp = pattern

        tmp = self._format_time(tmp)

        matches = re.findall("(%-?[0]?[0-9]?[AcmNnpPrtT])", tmp)
        for m in matches:
            if "A" in m:
                tmp = self._format_field(tmp, m, ovs.util.PROGRAM_NAME)
            elif "c" in m:
                tmp = self._format_field(tmp, m, self.name)
            elif "m" in m:
                tmp = self._format_field(tmp, m, message)
            elif "N" in m:
                tmp = self._format_field(tmp, m, str(msg_num))
            elif "n" in m:
                tmp = re.sub(m, "\n", tmp)
            elif "p" in m:
                tmp = self._format_field(tmp, m, level.upper())
            elif "P" in m:
                self._format_field(tmp, m, str(os.getpid()))
            elif "r" in m:
                now = datetime.datetime.utcnow()
                delta = now - self.__start_time
                ms = delta.microseconds / 1000
                tmp = self._format_field(tmp, m, str(ms))
            elif "t" in m:
                subprogram = threading.currentThread().getName()
                if subprogram == "MainThread":
                    subprogram = "main"
                tmp = self._format_field(tmp, m, subprogram)
            elif "T" in m:
                subprogram = threading.currentThread().getName()
                if not subprogram == "MainThread":
                    subprogram = "({})".format(subprogram)
                else:
                    subprogram = ""
                tmp = self._format_field(tmp, m, subprogram)
        return tmp.strip()

    def _format_field(self, tmp, match, replace):
        formatting = re.compile("^%(0)?([1-9])?")
        matches = formatting.match(match)
        # Do we need to apply padding?
        if not matches.group(1) and replace != "":
            replace = replace.center(len(replace) + 2)
        # Does the field have a minimum width
        if matches.group(2):
            min_width = int(matches.group(2))
            if len(replace) < min_width:
                replace = replace.center(min_width)
        return re.sub(match, replace.replace('\\', r'\\'), tmp)

    def _format_time(self, tmp):
        date_regex = re.compile(r'(%(0?[1-9]?[dD])(\{(.*)\})?)')
        match = date_regex.search(tmp)

        if match is None:
            return tmp

        # UTC date or Local TZ?
        if match.group(2) == "d":
            now = datetime.datetime.now()
        elif match.group(2) == "D":
            now = datetime.datetime.utcnow()

        # Custom format or ISO format?
        if match.group(3):
            time = datetime.date.strftime(now, match.group(4))
            try:
                i = len(re.search("#+", match.group(4)).group(0))
                msec = '{0:0>{i}.{i}}'.format(str(now.microsecond / 1000), i=i)
                time = re.sub('#+', msec, time)
            except AttributeError:
                pass
        else:
            time = datetime.datetime.isoformat(now.replace(microsecond=0))

        return self._format_field(tmp, match.group(1), time)

    def emer(self, message, **kwargs):
        self.__log("EMER", message, **kwargs)

    def err(self, message, **kwargs):
        self.__log("ERR", message, **kwargs)

    def warn(self, message, **kwargs):
        self.__log("WARN", message, **kwargs)

    def info(self, message, **kwargs):
        self.__log("INFO", message, **kwargs)

    def dbg(self, message, **kwargs):
        self.__log("DBG", message, **kwargs)

    def __is_enabled(self, level):
        level = LEVELS.get(level.lower(), logging.DEBUG)
        for f, f_level in Vlog.__mfl[self.name].items():
            f_level = LEVELS.get(f_level, logging.CRITICAL)
            if level >= f_level:
                return True
        return False

    def emer_is_enabled(self):
        return self.__is_enabled("EMER")

    def err_is_enabled(self):
        return self.__is_enabled("ERR")

    def warn_is_enabled(self):
        return self.__is_enabled("WARN")

    def info_is_enabled(self):
        return self.__is_enabled("INFO")

    def dbg_is_enabled(self):
        return self.__is_enabled("DBG")

    def exception(self, message):
        """Logs 'message' at ERR log level.  Includes a backtrace when in
        exception context."""
        self.err(message, exc_info=True)

    @staticmethod
    def init(log_file=None):
        """Intializes the Vlog module.  Causes Vlog to write to 'log_file' if
        not None.  Should be called after all Vlog objects have been created.
        No logging will occur until this function is called."""

        if Vlog.__inited:
            return

        Vlog.__inited = True
        Vlog.__start_time = datetime.datetime.utcnow()
        logging.raiseExceptions = False
        Vlog.__log_file = log_file
        for f in DESTINATIONS:
            logger = logging.getLogger(f)
            logger.setLevel(logging.DEBUG)

            try:
                if f == "console":
                    logger.addHandler(logging.StreamHandler(sys.stderr))
                elif f == "syslog":
                    Vlog.add_syslog_handler()
                elif f == "file" and Vlog.__log_file:
                    Vlog.__file_handler = logging.FileHandler(Vlog.__log_file)
                    logger.addHandler(Vlog.__file_handler)
            except (IOError, socket.error):
                logger.disabled = True

        ovs.unixctl.command_register("vlog/reopen", "", 0, 0,
                                     Vlog._unixctl_vlog_reopen, None)
        ovs.unixctl.command_register("vlog/close", "", 0, 0,
                                     Vlog._unixctl_vlog_close, None)
        try:
            # Windows limitation on Python 2, sys.maxsize is a long number
            # on 64 bits environments, while sys.maxint is the maximum int
            # number. Python 3 works as expected.
            maxsize_int = sys.maxint
        except AttributeError:
            maxsize_int = sys.maxsize
        ovs.unixctl.command_register("vlog/set", "spec", 1, maxsize_int,
                                     Vlog._unixctl_vlog_set, None)
        ovs.unixctl.command_register("vlog/list", "", 0, 0,
                                     Vlog._unixctl_vlog_list, None)

    @staticmethod
    def set_level(module, destination, level):
        """ Sets the log level of the 'module'-'destination' tuple to 'level'.
        All three arguments are strings which are interpreted the same as
        arguments to the --verbose flag.  Should be called after all Vlog
        objects have already been created."""

        module = module.lower()
        destination = destination.lower()
        level = level.lower()

        if destination != "any" and destination not in DESTINATIONS:
            return

        if module != "any" and module not in Vlog.__mfl:
            return

        if level not in LEVELS:
            return

        if module == "any":
            modules = list(Vlog.__mfl.keys())
        else:
            modules = [module]

        if destination == "any":
            destinations = list(DESTINATIONS.keys())
        else:
            destinations = [destination]

        for m in modules:
            for f in destinations:
                Vlog.__mfl[m][f] = level

    @staticmethod
    def set_pattern(destination, pattern):
        """ Sets the log pattern of the 'destination' to 'pattern' """
        destination = destination.lower()
        Vlog.__log_patterns[destination] = pattern

    @staticmethod
    def add_syslog_handler(facility=None):
        global syslog_facility, syslog_handler

        # If handler is already added and there is no change in 'facility',
        # there is nothing to do.
        if (not facility or facility == syslog_facility) and syslog_handler:
            return

        logger = logging.getLogger('syslog')
        # Disable the logger if the "null" syslog method requested
        # by environment.
        if os.environ.get('OVS_SYSLOG_METHOD') == "null":
            logger.disabled = True
            return

        if facility is None:
            facility = syslog_facility

        new_handler = logging.handlers.SysLogHandler(address="/dev/log",
                                                     facility=facility)

        if syslog_handler:
            logger.removeHandler(syslog_handler)

        syslog_handler = new_handler
        syslog_facility = facility

        logger.addHandler(syslog_handler)
        return

    @staticmethod
    def set_levels_from_string(s):
        module = None
        level = None
        destination = None

        words = re.split('[ :]', s)
        if words[0] == "pattern":
            try:
                if words[1] in DESTINATIONS and words[2]:
                    segments = [words[i] for i in range(2, len(words))]
                    pattern = "".join(segments)
                    Vlog.set_pattern(words[1], pattern)
                    return
                else:
                    return "Destination %s does not exist" % words[1]
            except IndexError:
                return "Please supply a valid pattern and destination"
        elif words[0] == "FACILITY":
            if words[1] in FACILITIES:
                try:
                    Vlog.add_syslog_handler(words[1])
                except (IOError, socket.error):
                    logger = logging.getLogger('syslog')
                    logger.disabled = True
                return
            else:
                return "Facility %s is invalid" % words[1]

        for word in [w.lower() for w in words]:
            if word == "any":
                pass
            elif word in DESTINATIONS:
                if destination:
                    return "cannot specify multiple destinations"
                destination = word
            elif word in LEVELS:
                if level:
                    return "cannot specify multiple levels"
                level = word
            elif word in Vlog.__mfl:
                if module:
                    return "cannot specify multiple modules"
                module = word
            else:
                return "no destination, level, or module \"%s\"" % word

        Vlog.set_level(module or "any", destination or "any", level or "any")

    @staticmethod
    def get_levels():
        lines = ["                 console    syslog    file\n",
                 "                 -------    ------    ------\n"]
        lines.extend(sorted(["%-16s  %4s       %4s       %4s\n"
                             % (m,
                                Vlog.__mfl[m]["console"],
                                Vlog.__mfl[m]["syslog"],
                                Vlog.__mfl[m]["file"]) for m in Vlog.__mfl]))
        return ''.join(lines)

    @staticmethod
    def reopen_log_file():
        """Closes and then attempts to re-open the current log file.  (This is
        useful just after log rotation, to ensure that the new log file starts
        being used.)"""

        if Vlog.__log_file:
            logger = logging.getLogger("file")
            logger.removeHandler(Vlog.__file_handler)
            Vlog.__file_handler = logging.FileHandler(Vlog.__log_file)
            logger.addHandler(Vlog.__file_handler)

    @staticmethod
    def close_log_file():
        """Closes the current log file. (This is useful on Windows, to ensure
        that a reference to the file is not kept by the daemon in case of
        detach.)"""
        if Vlog.__log_file:
            logger = logging.getLogger("file")
            logger.removeHandler(Vlog.__file_handler)
            Vlog.__file_handler.close()

    @staticmethod
    def _unixctl_vlog_reopen(conn, unused_argv, unused_aux):
        if Vlog.__log_file:
            Vlog.reopen_log_file()
            conn.reply(None)
        else:
            conn.reply("Logging to file not configured")

    @staticmethod
    def _unixctl_vlog_close(conn, unused_argv, unused_aux):
        if Vlog.__log_file:
            if sys.platform != 'win32':
                logger = logging.getLogger("file")
                logger.removeHandler(Vlog.__file_handler)
            else:
                Vlog.close_log_file()
        conn.reply(None)

    @staticmethod
    def _unixctl_vlog_set(conn, argv, unused_aux):
        for arg in argv:
            msg = Vlog.set_levels_from_string(arg)
            if msg:
                conn.reply(msg)
                return
        conn.reply(None)

    @staticmethod
    def _unixctl_vlog_list(conn, unused_argv, unused_aux):
        conn.reply(Vlog.get_levels())


def add_args(parser):
    """Adds vlog related options to 'parser', an ArgumentParser object.  The
    resulting arguments parsed by 'parser' should be passed to handle_args."""

    group = parser.add_argument_group(title="Logging Options")
    group.add_argument("--log-file", nargs="?", const="default",
                       help="Enables logging to a file.  Default log file"
                       " is used if LOG_FILE is omitted.")
    group.add_argument("-v", "--verbose", nargs="*",
                       help="Sets logging levels, see ovs-vswitchd(8)."
                       "  Defaults to dbg.")


def handle_args(args):
    """ Handles command line arguments ('args') parsed by an ArgumentParser.
    The ArgumentParser should have been primed by add_args().  Also takes care
    of initializing the Vlog module."""

    log_file = args.log_file
    if log_file == "default":
        log_file = "%s/%s.log" % (ovs.dirs.LOGDIR, ovs.util.PROGRAM_NAME)

    if args.verbose is None:
        args.verbose = []
    elif args.verbose == []:
        args.verbose = ["any:any:dbg"]

    for verbose in args.verbose:
        msg = Vlog.set_levels_from_string(verbose)
        if msg:
            ovs.util.ovs_fatal(0, "processing \"%s\": %s" % (verbose, msg))

    Vlog.init(log_file)



# Copyright (c) 2009, 2010, 2011 Nicira, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at:
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import ovs.json


class Error(Exception):
    def __init__(self, msg, json=None, tag=None):
        self.msg = msg
        self.json = json
        if tag is None:
            if json is None:
                self.tag = "ovsdb error"
            else:
                self.tag = "syntax error"
        else:
            self.tag = tag

        # Compose message.
        syntax = ""
        if self.json is not None:
            syntax = 'syntax "%s": ' % ovs.json.to_string(self.json)
        Exception.__init__(self, "%s%s: %s" % (syntax, self.tag, self.msg))



# Copyright (c) 2023 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at:
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Flow formatting framework.

This file defines a simple flow formatting framework. It's comprised of 3
classes: FlowStyle, FlowFormatter and FlowBuffer.

The FlowStyle arranges opaque style objects in a dictionary that can be queried
to determine what style a particular key-value should be formatted with.
That way, a particular implementation can represent its style using their own
object.

The FlowBuffer is an abstract class and must be derived by particular
implementations. It should know how to append parts of a flow using a style.
Only here the type of the style is relevant.

When asked to format a flow, the FlowFormatter will determine which style
the flow must be formatted with and call FlowBuffer functions with each part
of the flow and their corresponding style.
"""


class FlowStyle:
    """A FlowStyle determines the KVStyle to use for each key value in a flow.

    Styles are internally represented by a dictionary.
    In order to determine the style for a "key", the following items in the
    dictionary are fetched:
        - key.highlighted.{key} (if key is found in hightlighted)
        - key.highlighted (if key is found in hightlighted)
        - key.{key}
        - key
        - default

    In order to determine the style for a "value", the following items in the
    dictionary are fetched:
        - value.highlighted.{key} (if key is found in hightlighted)
        - value.highlighted.type{value.__class__.__name__}
        - value.highlighted
        (if key is found in hightlighted)
        - value.{key}
        - value.type.{value.__class__.__name__}
        - value
        - default

    The actual type of the style object stored for each item above is opaque
    to this class and it depends on the particular FlowFormatter child class
    that will handle them. Even callables can be stored, if so they will be
    called with the value of the field that is to be formatted and the return
    object will be used as style.

    Additionally, the following style items can be defined:
        - delim: for delimiters
        - delim.highlighted: for delimiters of highlighted key-values
    """

    def __init__(self, initial=None):
        self._styles = initial if initial is not None else dict()

    def __len__(self):
        return len(self._styles)

    def set_flag_style(self, kvstyle):
        self._styles["flag"] = kvstyle

    def set_delim_style(self, kvstyle, highlighted=False):
        if highlighted:
            self._styles["delim.highlighted"] = kvstyle
        else:
            self._styles["delim"] = kvstyle

    def set_default_key_style(self, kvstyle, highlighted=False):
        if highlighted:
            self._styles["key.highlighted"] = kvstyle
        else:
            self._styles["key"] = kvstyle

    def set_default_value_style(self, kvstyle, highlighted=False):
        if highlighted:
            self._styles["value.highlighted"] = kvstyle
        else:
            self._styles["value"] = kvstyle

    def set_key_style(self, key, kvstyle, highlighted=False):
        if highlighted:
            self._styles["key.highlighted.{}".format(key)] = kvstyle
        else:
            self._styles["key.{}".format(key)] = kvstyle

    def set_value_style(self, key, kvstyle, highlighted=None):
        if highlighted:
            self._styles["value.highlighted.{}".format(key)] = kvstyle
        else:
            self._styles["value.{}".format(key)] = kvstyle

    def set_value_type_style(self, name, kvstyle, highlighted=None):
        if highlighted:
            self._styles["value.highlighted.type.{}".format(name)] = kvstyle
        else:
            self._styles["value.type.{}".format(name)] = kvstyle

    def get(self, key):
        return self._styles.get(key)

    def get_delim_style(self, highlighted=False):
        delim_style_lookup = ["delim.highlighted"] if highlighted else []
        delim_style_lookup.extend(["delim", "default"])
        return next(
            (
                self._styles.get(s)
                for s in delim_style_lookup
                if self._styles.get(s)
            ),
            None,
        )

    def get_flag_style(self):
        return self._styles.get("flag") or self._styles.get("default")

    def get_key_style(self, kv, highlighted=False):
        key = kv.key

        key_style_lookup = (
            ["key.highlighted.%s" % key, "key.highlighted"]
            if highlighted
            else []
        )
        key_style_lookup.extend(["key.%s" % key, "key", "default"])

        style = next(
            (
                self._styles.get(s)
                for s in key_style_lookup
                if self._styles.get(s)
            ),
            None,
        )
        if callable(style):
            return style(kv.meta.kstring)
        return style

    def get_value_style(self, kv, highlighted=False):
        key = kv.key
        value_type = kv.value.__class__.__name__.lower()
        value_style_lookup = (
            [
                "value.highlighted.%s" % key,
                "value.highlighted.type.%s" % value_type,
                "value.highlighted",
            ]
            if highlighted
            else []
        )
        value_style_lookup.extend(
            [
                "value.%s" % key,
                "value.type.%s" % value_type,
                "value",
                "default",
            ]
        )

        style = next(
            (
                self._styles.get(s)
                for s in value_style_lookup
                if self._styles.get(s)
            ),
            None,
        )
        if callable(style):
            return style(kv.meta.vstring)
        return style


class FlowFormatter:
    """FlowFormatter is a base class for Flow Formatters."""

    def __init__(self):
        self._highlighted = list()

    def _style_from_opts(self, opts, opts_key, style_constructor):
        """Create style object from options.

        Args:
            opts (dict): Options dictionary
            opts_key (str): The options style key to extract
                (e.g: console or html)
            style_constructor(callable): A callable that creates a derived
                style object
        """
        if not opts or not opts.get("style"):
            return None

        section_name = ".".join(["styles", opts.get("style")])
        if section_name not in opts.get("config").sections():
            return None

        config = opts.get("config")[section_name]
        style = {}
        for key in config:
            (_, console, style_full_key) = key.partition(opts_key + ".")
            if not console:
                continue

            (style_key, _, prop) = style_full_key.rpartition(".")
            if not prop or not style_key:
                raise Exception("malformed style config: {}".format(key))

            if not style.get(style_key):
                style[style_key] = {}
            style[style_key][prop] = config[key]

        return FlowStyle({k: style_constructor(**v) for k, v in style.items()})

    def format_flow(self, buf, flow, style_obj=None, highlighted=None,
                    omitted=None):
        """Formats the flow into the provided buffer.

        Args:
            buf (FlowBuffer): the flow buffer to append to
            flow (ovs_dbg.OFPFlow): the flow to format
            style_obj (FlowStyle): Optional; style to use
            highlighted (list): Optional; list of KeyValues to highlight
            omitted (list): Optional; dict of keys to omit indexed by section
                name.
        """
        last_printed_pos = 0
        first = True

        if style_obj or omitted:
            style_obj = style_obj or FlowStyle()
            for section in sorted(flow.sections, key=lambda x: x.pos):
                section_omitted = (omitted or {}).get(section.name)
                if isinstance(section_omitted, str) and \
                   section_omitted == "all":
                    last_printed_pos += section.pos + len(section.string)
                    continue

                # Do not print leading extra strings (e.g: spaces and commas)
                # if it's the first section that gets printed.
                if not first:
                    buf.append_extra(
                        flow.orig[last_printed_pos : section.pos],
                        style=style_obj.get("default"),
                    )

                self.format_kv_list(
                    buf, section.data, section.string, style_obj, highlighted,
                    section_omitted
                )
                last_printed_pos = section.pos + len(section.string)
                first = False
        else:
            # Don't pay the cost of formatting each section one by one.
            buf.append_extra(flow.orig.strip(), None)

    def format_kv_list(self, buf, kv_list, full_str, style_obj, highlighted,
                      omitted=None):
        """Format a KeyValue List.

        Args:
            buf (FlowBuffer): a FlowBuffer to append formatted KeyValues to
            kv_list (list[KeyValue]: the KeyValue list to format
            full_str (str): the full string containing all k-v
            style_obj (FlowStyle): a FlowStyle object to use
            highlighted (list): Optional; list of KeyValues to highlight
            highlighted (list): Optional; list of KeyValues to highlight
            omitted (list): Optional; list of keys to omit
        """
        for i, kv in enumerate(kv_list):
            key_omitted = kv.key in omitted if omitted else False
            written = self.format_kv(
                buf, kv, style_obj=style_obj, highlighted=highlighted,
                omitted=key_omitted
            )

            end = (
                kv_list[i + 1].meta.kpos
                if i < (len(kv_list) - 1)
                else len(full_str)
            )

            buf.append_extra(
                full_str[(kv.meta.kpos + written) : end].rstrip("\n\r"),
                style=style_obj.get("default"),
            )

    def format_kv(self, buf, kv, style_obj, highlighted=None, omitted=False):
        """Format a KeyValue

        A formatted keyvalue has the following parts:
            {key}{delim}{value}[{delim}]

        Args:
            buf (FlowBuffer): buffer to append the KeyValue to
            kv (KeyValue): The KeyValue to print
            style_obj (FlowStyle): The style object to use
            highlighted (list): Optional; list of KeyValues to highlight
            omitted(boolean): Whether the value shall be omitted.

        Returns the number of printed characters.
        """
        ret = 0
        key = kv.meta.kstring
        is_highlighted = (
            key in [k.key for k in highlighted] if highlighted else False
        )

        key_style = style_obj.get_key_style(kv, is_highlighted)
        buf.append_key(kv, key_style)  # format value
        ret += len(key)

        if not kv.meta.vstring:
            return ret

        if kv.meta.delim not in ("\n", "\t", "\r", ""):
            buf.append_delim(kv, style_obj.get_delim_style(is_highlighted))
            ret += len(kv.meta.delim)

        if omitted:
            buf.append_value_omitted(kv)
            ret += len(kv.meta.vstring)

        else:
            value_style = style_obj.get_value_style(kv, is_highlighted)
            buf.append_value(kv, value_style)  # format value
            ret += len(kv.meta.vstring)

        if kv.meta.end_delim:
            buf.append_end_delim(kv, style_obj.get_delim_style(is_highlighted))
            ret += len(kv.meta.end_delim)

        return ret


class FlowBuffer:
    """A FlowBuffer is a base class for format buffers.

    Childs must implement the following methods:
        append_key(self, kv, style)
        append_value(self, kv, style)
        append_delim(self, delim, style)
        append_end_delim(self, delim, style)
        append_extra(self, extra, style)
    """

    def append_key(self, kv, style):
        """Append a key.
        Args:
            kv (KeyValue): the KeyValue instance to append
            style (Any): the style to use
        """
        raise NotImplementedError

    def append_delim(self, kv, style):
        """Append a delimiter.
        Args:
            kv (KeyValue): the KeyValue instance to append
            style (Any): the style to use
        """
        raise NotImplementedError

    def append_end_delim(self, kv, style):
        """Append an end delimiter.
        Args:
            kv (KeyValue): the KeyValue instance to append
            style (Any): the style to use
        """
        raise NotImplementedError

    def append_value(self, kv, style):
        """Append a value.
        Args:
            kv (KeyValue): the KeyValue instance to append
            style (Any): the style to use
        """
        raise NotImplementedError

    def append_value_omitted(self, kv):
        """Append an omitted value.
        Args:
            kv (KeyValue): the KeyValue instance to append
        """
        raise NotImplementedError

    def append_extra(self, extra, style):
        """Append extra string.
        Args:
            kv (KeyValue): the KeyValue instance to append
            style (Any): the style to use
        """
        raise NotImplementedError



# Copyright (c) 2023 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at:
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ovs.flowviz.html_format import HTMLBuffer, HTMLFormatter
from ovs.flowviz.odp.tree import FlowTree
from ovs.flowviz.process import FileProcessor


class HTMLTree:
    """Class capable of printing a FlowTree in HTML."""

    BODY_STYLE = """
    <style>
    body {{
        background-color: {bg};
        color: {fg};
    }}
    </style>"""

    STYLE = """
    <style>
    .recirc {
        font-weight: bold;
        font-family: monospace;
        font-size: 1.1rem;
        border: 3px solid #ccc;
        width: fit-content;
        block-size: fit-content;
    }

    .block-matches {
        font-family: monospace;
        border: 1px solid #ccc;
        width: fit-content;
        block-size: fit-content;
        margin-left: 1em;
        padding: 4px;
    }

    .block-actions {
        font-family: monospace;
        border: 2px solid #ccc;
        width: fit-content;
        block-size: fit-content;
        margin-top: 0.5em;
        margin-bottom : 2em;
        margin-left: 1em;
        padding: 4px;
    }

    /* List styling */
    ul il {
        list-style-type: disc;
    }

    .flowlist > li::marker {
        list-style: disc;
    }

    .actions > li::marker  {
        content: "\\21B3";
        font-size: 1.5em;
    }

    /* Caret styling */
    .caret {
        cursor: pointer;
        user-select: none;
    }

    .caret::before {
        content: "\\25B6";
        font-size: 1.2em;
        display: inline-block;
        margin-right: 5px;cursor: pointer;
    }

    /* Rotate the caret/arrow icon when clicked on. */
    .caret-down::before {
        transform: rotate(90deg);
    }

    .nested {
        display: none;
    }

    .active {
        display: block;
    }

    .focused {
        border: 2px solid #0008ff;
    }
    </style>
    """  # noqa: E501

    SCRIPT = """
    <script>
    var caret = document.getElementsByClassName("caret");
    var blocks = document.getElementsByClassName("block-matches");
    var i;

    for (i = 0; i < caret.length; i++) {
        caret[i].addEventListener("click", function() {
            this.parentElement.querySelector(".nested").classList.toggle("active");
            this.classList.toggle("caret-down");
        });
    }

    // Set focus to a flow block, expanding all parent elements.
    function setFocus(targetId) {
        var target = document.getElementById(targetId);
        var others = document.getElementsByClassName("focused");
        var i;
        for (i = 0; i < others.length; i++) {
            others[i].classList.remove("focused");
        }
        if (target) {
            var element = target;
            while (element !== null) {
            if (element.classList.contains("nested")) {
                element.classList.add("active");
            }
            if (element.previousElementSibling && element.previousElementSibling.classList.contains("caret")) {
                element.previousElementSibling.classList.add("caret-down");
            }
            element = element.parentElement;
            }
        }
        target.classList.toggle("focused");
    }

    window.addEventListener('hashchange', function () {
        var targetId = window.location.hash.substring(1);
        setFocus(targetId);
    });
    </script>
    """  # noqa: E501

    def __init__(self, name, flowtree, opts):
        self.name = name.replace(" ", "_")
        self.tree = flowtree
        self.opts = opts
        self.formatter = HTMLFormatter(opts)

    @classmethod
    def head(cls):
        html = "<head>"
        html += cls.STYLE
        html += "</head>"

        return html

    @classmethod
    def begin_body(cls, opts):
        style = HTMLFormatter(opts).style
        bg = (
            style.get("background").color
            if style.get("background")
            else "#f0f0f0"
        )
        fg = style.get("default").color if style.get("default") else "black"
        return cls.BODY_STYLE.format(bg=bg, fg=fg)

    @classmethod
    def end_body(cls):
        return cls.SCRIPT

    def format(self):
        html_obj = f"<div id=flow_list-{self.name}>"

        html_obj += '<ul class="active flowlist">'
        for in_port in sorted(self.tree.recirc_nodes[0].keys()):
            node = self.tree.recirc_nodes[0][in_port]
            if node.visible:
                html_obj += "<li>"
                html_obj += self.format_recirc_node(node)
                html_obj += "</li>"

        html_obj += "</ul>"
        html_obj += "</div>"
        return html_obj

    def format_recirc_node(self, node):
        html_obj = '<div class="recirc">'
        html_obj += "[recirc_id({}) in_port({})]".format(
            hex(node.recirc), node.in_port
        )
        html_obj += "</div>"

        html_obj += '<ul class="flowlist">'  # nested

        for block in node.visible_blocks():
            html_block = "<li>"
            html_block += self.format_block(block)
            html_block += "</li>"
            html_obj += html_block

        html_obj += "</ul>"
        return html_obj

    def format_single_block(self, block):
        block_id = "block_{}".format(block.flows[0].flow.id)
        html_obj = f'<div id="{block_id}" class="block-matches">'

        omit_first = {
            "actions": "all",
        }
        omit_rest = {
            "actions": "all",
            "match": [kv.key for _, kv in block.equal_match],
        }

        for i, flow in enumerate(filter(lambda x: x.visible, block.flows)):
            html_obj += '<div class="flow">'

            omit = omit_rest if i > 0 else omit_first
            buf = HTMLBuffer()
            hl = None
            if self.opts.get("highlight"):
                result = self.opts.get("highlight").evaluate(flow.flow)
                if result:
                    hl = result.kv

            self.formatter.format_flow(buf, flow.flow, hl, omitted=omit)
            html_obj += buf.text
            html_obj += "</div>"

        html_obj += "</div>"  # Match list.

        html_obj += '<ul class="actions"><li>'
        html_obj += "<div>"  # Match list.
        if block.next_recirc_nodes:
            html_obj += '<div class="caret block-actions">'
        else:
            html_obj += '<div class="block-actions">'

        omit = {
            "match": "all",
            "info": "all",
            "ufid": "all",
            "dp_extra_info": "all",
        }
        buf = HTMLBuffer()
        buf.append_extra("actions: ", None)

        hl = None
        if self.opts.get("highlight"):
            result = self.opts.get("highlight").evaluate(block.flows[0].flow)
            if result:
                hl = result.kv

        self.formatter.format_flow(buf, block.flows[0].flow, hl, omitted=omit)
        html_obj += buf.text
        html_obj += "</div>"
        return html_obj

    def format_block(self, block):
        html_obj = self.format_single_block(block)

        html_obj += '<ul class="nested recirclist">'
        for node in block.next_recirc_nodes:
            if node.visible:
                html_obj += "<li>"
                html_obj += self.format_recirc_node(node)
                html_obj += "</li>"
        html_obj += "</ul>"
        html_obj += "</li>"
        html_obj += "</ul>"
        return html_obj


class HTMLTreeProcessor(FileProcessor):
    def __init__(self, opts):
        super().__init__(opts, "odp")
        self.trees = {}
        self.opts = opts
        self.tree = None
        self.curr_file = ""

    def start_file(self, name, filename):
        self.tree = FlowTree()
        self.curr_file = name

    def start_thread(self, name):
        if not self.tree:
            self.tree = FlowTree()

    def stop_thread(self, name):
        full_name = self.curr_file + f" ({name})"
        if self.tree:
            self.trees[full_name] = self.tree
            self.tree = None

    def process_flow(self, flow, name):
        self.tree.add(flow, self.opts.get("filter"))

    def process(self):
        super().process(False)

    def stop_file(self, name, filename):
        if self.tree:
            self.trees[name] = self.tree
            self.tree = None

    def print(self):
        html_obj = "<html>"
        html_obj += "<head>"
        html_obj += HTMLTree.head()
        html_obj += "</head>"

        html_obj += "<body>"
        html_obj += HTMLTree.begin_body(self.opts)

        for name, tree in self.trees.items():
            tree.build()
            html_tree = HTMLTree(name, tree, self.opts)
            html_obj += "<div>"
            html_obj += "<h2>{}</h2>".format(name)
            html_obj += html_tree.format()
            html_obj += "</div>"

        html_obj += HTMLTree.end_body()
        html_obj += "</body>"
        html_obj += "</html>"
        print(html_obj)



# Copyright (c) 2023 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at:
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import json
import click

from ovs.flow.decoders import FlowEncoder
from ovs.flow.odp import ODPFlow
from ovs.flow.ofp import OFPFlow

from ovs.flowviz.console import (
    ConsoleFormatter,
    default_highlight,
    file_header,
    heat_pallete,
)
from ovs.flowviz.format import FlowStyle


class FileProcessor(object):
    """Base class for file-based Flow processing. It is able to create flows
    from strings found in a file (or stdin).

    The process of parsing the flows is extendable in many ways by deriving
    this class.

    When process() is called, the base class will:
        - call self.start_file() for each new file that get's processed
        - call self.start_thread() for each thread (Datapath flow only)
        - apply the filter defined in opts if provided (can be optionally
            disabled)
        - call self.process_flow() for after the flow has been filtered
        - call self.stop_thread() after the thread block has been processed
            (Datapath flow only)
        - call self.stop_file() after the file has been processed entirely

    In the case of stdin, the filename and file alias is 'stdin'.

    Args:
        opts (dict): Options dictionary
        flow_type (str): ["ofp", "odp"]
    """

    def __init__(self, opts, flow_type):
        self.opts = opts
        assert flow_type in ["ofp", "odp"]
        self.flow_type = flow_type
        self.current_thread = None

    # Methods that must be implemented by derived classes.
    def init(self):
        """Called before the flow processing begins."""
        pass

    def start_file(self, alias, filename):
        """Called before the processing of a file begins.
        Args:
            alias(str): The alias name of the filename
            filename(str): The filename string
        """
        pass

    def start_thread(self, thread):
        """Called before the processing of a thread block begins.
        Args:
            thread(str): The thread name ("main" or "pmd at cpu core $N")
        """
        raise NotImplementedError

    def process_flow(self, flow, name):
        """Called for built flow (after filtering).
        Args:
            flow(Flow): The OpenFlow or Datapath flow.
            name(str): The name of the file from which the flow comes
        """
        raise NotImplementedError

    def stop_thread(self, thread):
        """Called after the processing of a thread ends.
        Args:
            thread(str): The thread name ("main" or "pmd at cpu core $N")
        """
        raise NotImplementedError

    def stop_file(self, alias, filename):
        """Called after the processing of a file ends.
        Args:
            alias(str): The alias name of the filename
            filename(str): The filename string
        """
        pass

    def end(self):
        """Called after the processing ends."""
        pass

    def process_line(self, line, idx):
        if self.flow_type == "odp":
            next_thread = self.current_thread
            if line.startswith("flow-dump from the main thread"):
                next_thread = "main"
            elif line.startswith("flow-dump from pmd on cpu core"):
                next_thread = line.removeprefix("flow-dump from ").strip("\n")

            if next_thread != self.current_thread:
                if self.current_thread:
                    self.stop_thread(self.current_thread)
                self.start_thread(next_thread)
                self.current_thread = next_thread
                return None

            return ODPFlow(line, idx)

        elif self.flow_type == "ofp":
            # Skip strings commonly found in OpenFlow flow dumps.
            if " reply " in line:
                return None

            return OFPFlow(line, idx)

    def process(self, do_filter=True):
        idx = 0
        filenames = self.opts.get("filename")
        filt = self.opts.get("filter") if do_filter else None
        self.init()
        if filenames:
            for alias, filename in filenames:
                try:
                    with open(filename) as f:
                        self.start_file(alias, filename)
                        for line in f:
                            flow = self.process_line(line, idx)
                            idx += 1
                            if not flow or (filt and not filt.evaluate(flow)):
                                continue
                            self.process_flow(flow, alias)
                        if self.current_thread:
                            self.stop_thread(self.current_thread)
                        self.stop_file(alias, filename)
                except IOError as e:
                    raise click.BadParameter(
                        "Failed to read from file {} ({}): {}".format(
                            filename, e.errno, e.strerror
                        )
                    )
        else:
            data = sys.stdin.read()
            self.start_file("stdin", "stdin")
            for line in data.split("\n"):
                line = line.strip()
                if line:
                    flow = self.process_line(line, idx)
                    idx += 1
                    if (
                        not flow
                        or not getattr(flow, "_sections", None)
                        or (filt and not filt.evaluate(flow))
                    ):
                        continue
                    self.process_flow(flow, "stdin")
            if self.current_thread:
                self.stop_thread(self.current_thread)
            self.stop_file("stdin", "stdin")
        self.end()


class JSONOpenFlowProcessor(FileProcessor):
    """A FileProcessor that prints OpenFlow flows in JSON format."""

    def __init__(self, opts):
        super().__init__(opts, "ofp")
        self.flows = dict()

    def start_file(self, name, filename):
        self.flows_list = list()

    def stop_file(self, name, filename):
        self.flows[name] = self.flows_list

    def process_flow(self, flow, name):
        self.flows_list.append(flow)

    def json_string(self):
        if len(self.flows.keys()) > 1:
            return json.dumps(
                [
                    {"name": name, "flows": [flow.dict() for flow in flows]}
                    for name, flows in self.flows.items()
                ],
                indent=4,
                cls=FlowEncoder,
            )
        return json.dumps(
            [flow.dict() for flow in self.flows_list],
            indent=4,
            cls=FlowEncoder,
        )


class JSONDatapathProcessor(FileProcessor):
    """A FileProcessor that prints Datapath flows in JSON format."""

    def __init__(self, opts):
        super().__init__(opts, "odp")
        self.data = {}
        self.thread = None
        self.file = None

    def start_file(self, name, filename):
        self.per_thread_flows = None
        self.flows_list = []

    def start_thread(self, name):
        if not self.per_thread_flows:
            self.per_thread_flows = {}

    def stop_thread(self, name):
        self.per_thread_flows[name] = self.flows_list

    def stop_file(self, name, filename):
        if self.per_thread_flows:
            self.data[name] = self.per_thread_flows
        else:
            self.data[name] = self.flows_list

    def process_flow(self, flow, name):
        self.flows_list.append(flow)

    def json_string(self):
        opts = {
            "indent": 4,
            "cls": FlowEncoder,
        }

        def thread_data(data):
            if isinstance(data, dict):
                return {
                    thread: [flow.dict() for flow in flows]
                    for thread, flows in data.items()
                }
            return [flow.dict() for flow in data]

        if len(self.data.keys()) > 1:
            jsondata = {}
            for file, file_data in self.data.items():
                jsondata[file] = thread_data(file_data)
            return json.dumps(jsondata, **opts)
        else:
            return json.dumps(
                thread_data(next(iter(self.data.values()))), **opts
            )


class ConsoleProcessor(FileProcessor):
    """A generic Console Processor that prints flows into the console"""

    def __init__(self, opts, flow_type, heat_map=[]):
        super().__init__(opts, flow_type)
        self.heat_map = heat_map
        self.console = ConsoleFormatter(opts)
        if not self.console.style and self.opts.get("highlight"):
            # Add some style to highlights or else they won't be seen.
            self.console.style = FlowStyle()
            self.console.style.set_default_value_style(
                default_highlight(), True
            )
            self.console.style.set_default_key_style(default_highlight(), True)

        self.flows = dict()  # Dict of flow-lists, one per file and thread.
        self.min_max = dict()  # Used for heat-map calculation.
        self.curr_file = None
        self.flows_list = None

    def _init_list(self):
        self.flows_list = list()
        if len(self.heat_map) > 0:
            self.min = [-1] * len(self.heat_map)
            self.max = [0] * len(self.heat_map)

    def _save_list(self, name):
        if self.flows_list:
            self.flows[name] = self.flows_list
            self.flows_list = None
            if len(self.heat_map) > 0:
                self.min_max[name] = (self.min, self.max)

    def start_file(self, name, filename):
        self._init_list()
        self.curr_file = name

    def start_thread(self, name):
        if not self.flows_list:
            self._init_list()

    def stop_thread(self, name):
        full_name = self.curr_file + f" ({name})"
        self._save_list(full_name)

    def stop_file(self, name, filename):
        self._save_list(name)

    def process_flow(self, flow, name):
        # Running calculation of min and max values for all the fields that
        # take place in the heatmap.
        for i, field in enumerate(self.heat_map):
            val = flow.info.get(field)
            if self.min[i] == -1 or val < self.min[i]:
                self.min[i] = val
            if val > self.max[i]:
                self.max[i] = val

        self.flows_list.append(flow)

    def print(self):
        for name, flows in self.flows.items():
            self.console.console.print("\n")
            self.console.console.print(file_header(name))

            if len(self.heat_map) > 0 and len(self.flows) > 0:
                for i, field in enumerate(self.heat_map):
                    (min_val, max_val) = self.min_max[name][i]
                    self.console.style.set_value_style(
                        field, heat_pallete(min_val, max_val)
                    )

            for flow in flows:
                high = None
                if self.opts.get("highlight"):
                    result = self.opts.get("highlight").evaluate(flow)
                    if result:
                        high = result.kv
                self.console.print_flow(flow, high)



# Copyright (c) 2023 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at:
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ovs.flowviz.html_format import HTMLBuffer, HTMLFormatter, HTMLStyle
from ovs.flowviz.process import (
    FileProcessor,
)


class HTMLProcessor(FileProcessor):
    """File processor that prints OpenFlow tables in HTML."""

    def __init__(self, opts):
        super().__init__(opts, "ofp")
        self.formatter = HTMLFormatter(self.opts)
        self.data = dict()

    def start_file(self, name, filename):
        self.tables = dict()

    def stop_file(self, name, filename):
        self.data[name] = self.tables

    def process_flow(self, flow, name):
        table = flow.info.get("table") or 0
        if not self.tables.get(table):
            self.tables[table] = list()
        self.tables[table].append(flow)

    def html(self):
        bg = (
            self.formatter.style.get("background").color
            if self.formatter.style.get("background")
            else "white"
        )
        fg = (
            self.formatter.style.get("default").color
            if self.formatter.style.get("default")
            else "black"
        )
        html_obj = """
        <style>
        body {{
            background-color: {bg};
            color: {fg};
        }}
        </style>""".format(
            bg=bg, fg=fg
        )
        for name, tables in self.data.items():
            name = name.replace(" ", "_")
            html_obj += "<h1>{}</h1>".format(name)
            html_obj += "<div id=flow_list>"
            for table, flows in tables.items():

                def anchor(x):
                    return "#table_%s_%s" % (name, x.value["table"])

                resubmit_style = self.formatter.style.get("value.resubmit")
                resubmit_color = resubmit_style.color if resubmit_style else fg

                self.formatter.style.set_value_style(
                    "resubmit",
                    HTMLStyle(
                        resubmit_color,
                        anchor_gen=anchor,
                    ),
                )
                html_obj += (
                    "<h2 id=table_{name}_{table}> Table {table}</h2>".format(
                        name=name, table=table
                    )
                )
                html_obj += "<ul id=table_{}_flow_list>".format(table)
                for flow in flows:
                    html_obj += "<li id=flow_{}>".format(flow.id)
                    highlighted = None
                    if self.opts.get("highlight"):
                        result = self.opts.get("highlight").evaluate(flow)
                        if result:
                            highlighted = result.kv
                    buf = HTMLBuffer()
                    self.formatter.format_flow(buf, flow, highlighted)
                    html_obj += buf.text
                    html_obj += "</li>"
                html_obj += "</ul>"
            html_obj += "</div>"

        return html_obj
